#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import json
import httplib
from urllib import urlencode, urlretrieve
import logging
import zipfile
import re
import shutil
from tempfile import gettempdir
from .config import Conf
from . import tools

C = Conf()


class PluginManager(object):

    _subSystems = {
        'admin': '/{admin}',
        'auth': '/auth',
        'availability': '/availability',
        'backup': '/backup/util/ui',
        'badges': '/badges',
        'block': '/blocks',
        'blog': '/blog',
        'cache': '/cache',
        'calendar': '/calendar',
        'cohort': '/cohort',
        'competency': '/competency',
        'course': '/course',
        'editor': '/lib/editor',
        'enrol': '/enrol',
        'files': '/files',
        'form': '/lib/form',
        'grades': '/grade',
        'grading': '/grade/grading',
        'group': '/group',
        'message': '/message',
        'mnet': '/mnet',
        'my': '/my',
        'notes': '/notes',
        'plagiarism': '/plagiarism',
        'portfolio': '/portfolio',
        'publish': '/course/publish',
        'question': '/question',
        'rating': '/rating',
        'register': '/{admin}/registration',
        'repository': '/repository',
        'rss': '/rss',
        'role': '/{admin}/roles',
        'search': '/search',
        'tag': '/tag',
        'user': '/user',
        'webservice': '/webservice'
    }

    _pluginTypesPath = {
        'antivirus': '/lib/antivirus',
        'availability': '/availability/condition',
        'qtype': '/question/type',
        'mod': '/mod',
        'auth': '/auth',
        'calendartype': '/calendar/type',
        'enrol': '/enrol',
        'message': '/message/output',
        'block': '/blocks',
        'filter': '/filter',
        'editor': '/lib/editor',
        'format': '/course/format',
        'profilefield': '/user/profile/field',
        'report': '/report',
        'coursereport': '/course/report',  # Must be after system reports.
        'gradeexport': '/grade/export',
        'gradeimport': '/grade/import',
        'gradereport': '/grade/report',
        'gradingform': '/grade/grading/form',
        'mnetservice': '/mnet/service',
        'webservice': '/webservice',
        'repository': '/repository',
        'portfolio': '/portfolio',
        'search': '/search/engine',
        'qbehaviour': '/question/behaviour',
        'qformat': '/question/format',
        'plagiarism': '/plagiarism',
        'tool': '/{admin}/tool',
        'cachestore': '/cache/stores',
        'cachelock': '/cache/locks',

        'theme': '/theme',
        'local': '/local'
    }
    _supportSubtypes = ['mod', 'editor', 'local', 'tool']

    @classmethod
    def extract(cls, f, plugin, M, override=False):
        """Extract a plugin zip file to the plugin directory of M"""

        if type(plugin) != PluginObject:
            raise ValueError('PluginObject expected')

        if not override and cls.hasPlugin(plugin, M):
            raise Exception('Plugin directory already exists')

        if not cls.validateZipFile(f, plugin.name):
            raise Exception('Invalid zip file')

        zp = zipfile.ZipFile(f)
        try:
            logging.info('Extracting plugin...')
            rootDir = os.path.commonprefix(zp.namelist())
            extractIn = cls.getTypeDirectory(plugin.t, M)
            zp.extractall(extractIn)
            if plugin.name != rootDir.rstrip('/'):
                orig = os.path.join(extractIn, rootDir).rstrip('/')
                dest = os.path.join(extractIn, plugin.name).rstrip('/')

                # Merge directories
                for src_dir, dirs, files in os.walk(orig):
                    dst_dir = src_dir.replace(orig, dest)
                    if not os.path.exists(dst_dir):
                        os.mkdir(dst_dir)
                    for file_ in files:
                        src_file = os.path.join(src_dir, file_)
                        dst_file = os.path.join(dst_dir, file_)
                        if os.path.exists(dst_file):
                            os.remove(dst_file)
                        shutil.move(src_file, dst_dir)

                shutil.rmtree(orig)

        except OSError:
            raise Exception('Error while extracting the files')

    @classmethod
    def getTypeAndName(cls, plugin):
        """Accepts a full plugin name 'mod_book' and returns the type and plugin name"""

        if plugin == 'moodle' or plugin == 'core' or plugin == '':
            return ('core', None)

        if not '_' in plugin:
            t = 'mod'
            name = plugin

        else:
            (t, name) = plugin.split('_', 1)
            if t == 'moodle':
                t = 'core'

        return (t, name)

    @classmethod
    def getSubsystems(cls):
        """Return the list of subsytems and their relative directory"""
        return cls._subSystems

    @classmethod
    def getSubsystemDirectory(cls, subsystem, M=None):
        """Return the subsystem directory, absolute if M is passed"""
        path = cls._subSystems.get(subsystem)
        if not path:
            raise ValueError('Unknown subsystem')

        if M:
            path = path.replace('{admin}', M.get('admin', 'admin'))
            path = os.path.join(M.get('path'), path.strip('/'))

        return path

    @classmethod
    def getSubsystemOrPluginFromPath(cls, path, M=None):
        """Get a subsystem from a path. Path should be relative to dirroot or M should be passed.

        This returns a tuple containing the name of the subsystem or plugin type, and the plugin name
        if we could resolve one.
        """

        subtypes = {}
        path = os.path.realpath(os.path.abspath(path))
        if M:
            path = '/' + path.replace(M.get('path'), '').strip('/')
            admindir = M.get('admin', 'admin')
            if path.startswith('/' + admindir):
                path = re.sub(r'^/%s' % admindir, '/{admin}', path)
            subtypes = cls.getSubtypes(M)
        path = '/' + path.lstrip('/')

        pluginOrSubsystem = None
        pluginName = None
        candidate = path
        head = True
        tail = None
        while head and head != '/' and not pluginOrSubsystem:
            # Check plugin types.
            if not pluginOrSubsystem:
                for k, v in cls._pluginTypesPath.iteritems():
                    if v == candidate:
                        pluginOrSubsystem = k
                        pluginName = tail
                        break

            # Check sub plugin types.
            if not pluginOrSubsystem:
                for k, v in subtypes.iteritems():
                    if v == candidate:
                        pluginOrSubsystem = k
                        pluginName = tail
                        break

            # Check subsystems.
            for k, v in cls._subSystems.iteritems():
                if v == candidate:
                    pluginOrSubsystem = k
                    break

            (head, tail) = os.path.split(candidate)
            candidate = head

        return (pluginOrSubsystem, pluginName)

    @classmethod
    def getSubtypes(cls, M):
        """Get the sub plugins declared in an instance"""
        regex = re.compile(r'\s*(?P<brackets>[\'"])(.*?)(?P=brackets)\s*=>\s*(?P=brackets)(.*?)(?P=brackets)')
        subtypes = {}
        for t in cls._supportSubtypes:
            path = cls.getTypeDirectory(t, M)
            dirs = os.listdir(path)
            for d in dirs:
                if not os.path.isdir(os.path.join(path, d)):
                    continue
                subpluginsfile = os.path.join(path, d, 'db', 'subplugins.php')
                if not os.path.isfile(subpluginsfile):
                    continue

                searchOpen = False
                f = open(subpluginsfile, 'r')
                for line in f:
                    if '$subplugins' in line:
                        searchOpen = True

                    if searchOpen:
                        search = regex.findall(line)
                        if search:
                            for match in search:
                                subtypes[match[1]] = '/' + match[2].replace('admin/', '{admin}/').lstrip('/')

                    # Exit when we find a semi-colon.
                    if searchOpen and ';' in line:
                        break

        return subtypes

    @classmethod
    def getTypeDirectory(cls, t, M=None):
        """Returns the path to the plugin type directory. If M is passed, the full path is returned."""
        path = cls._pluginTypesPath.get(t, False)
        if not path:
            if M:
                subtypes = cls.getSubtypes(M)
                path = subtypes.get(t, False)
            if not path:
                raise ValueError('Unknown plugin or subplugin type')

        if M:
            if t == 'theme':
                themedir = M.get('themedir', None)
                if themedir != None:
                    return themedir

            path = path.replace('{admin}', M.get('admin', 'admin'))
            path = os.path.join(M.get('path'), path.strip('/'))

        return path

    @classmethod
    def hasPlugin(cls, plugin, M):
        path = cls.getTypeDirectory(plugin.t, M)
        target = os.path.join(path, plugin.name)
        return os.path.isdir(target)

    @classmethod
    def validateZipFile(cls, f, name):
        zp = zipfile.ZipFile(f, 'r')

        # Checking that the content is all contained in one single directory
        rootDir = os.path.commonprefix(zp.namelist())
        if rootDir == '':
            return False
        return True

    @classmethod
    def deleteDirectoryTree(cls, plugin, M):
        directory = cls.getTypeDirectory(plugin.t, M)
        fullpath = directory + '/' + plugin.name
        if os.path.isdir(fullpath):
            shutil.rmtree(fullpath)


class PluginObject(object):

    component = None
    t = None
    name = None

    def __init__(self, component):
        self.component = component
        (self.t, self.name) = PluginManager.getTypeAndName(component)
        self.dlinfo = {}

    def getDownloadInfo(self, branch):
        if not self.dlinfo.get(branch, False):
            self.dlinfo[branch] = PluginRepository().info(self.component, branch)
        return self.dlinfo.get(branch, False)

    def getZip(self, branch, fileCache=None):
        dlinfo = self.getDownloadInfo(branch)
        if not dlinfo:
            return False
        return dlinfo.download(fileCache)


class PluginDownloadInfo(dict):

    def download(self, fileCache=None, cacheDir=C.get('dirs.mdk')):
        """Download a plugin"""

        if fileCache == None:
            fileCache = C.get('plugins.fileCache')

        dest = os.path.abspath(os.path.expanduser(os.path.join(cacheDir, 'plugins')))
        if not fileCache:
            dest = gettempdir()

        if not 'downloadurl' in self.keys():
            raise ValueError('Expecting the key downloadurl')
        elif not 'component' in self.keys():
            raise ValueError('Expecting the key component')
        elif not 'branch' in self.keys():
            raise ValueError('Expecting the key branch')

        dl = self.get('downloadurl')
        plugin = self.get('component')
        branch = self.get('branch')
        target = os.path.join(dest, '%s-%d.zip' % (plugin, branch))
        md5sum = self.get('downloadmd5')
        release = self.get('release', 'Unknown')

        if fileCache:
            if not os.path.isdir(dest):
                logging.debug('Creating directory %s' % (dest))
                tools.mkdir(dest, 0777)

            if os.path.isfile(target) and (md5sum == None or tools.md5file(target) == md5sum):
                logging.info('Found cached plugin file: %s' % (os.path.basename(target)))
                return target

        logging.info('Downloading %s (%s)' % (plugin, release))
        if logging.getLogger().level <= logging.INFO:
            urlretrieve(dl, target, tools.downloadProcessHook)
            # Force a new line after the hook display
            logging.info('')
        else:
            urlretrieve(dl, target)

        # Highly memory inefficient MD5 check
        if md5sum and tools.md5file(target) != md5sum:
            os.remove(target)
            logging.warning('Bad MD5 sum on downloaded file')
            return False

        return target


class PluginRepository(object):

    apiversion = '1.2'
    uri = '/api'
    host = 'download.moodle.org'
    ssl = True
    localRepository = None
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(PluginRepository, cls).__new__(cls, *args, **kwargs)
            cls._instance.localRepository = C.get('plugins.localRepository')
            cls._instance.localRepository = {} if cls._instance.localRepository == None else cls._instance.localRepository
        return cls._instance

    def info(self, plugin, branch):
        """Gets the download information of the plugin, branch is expected to be
        a whole integer, such as 25 for 2.5, etc...
        """

        if type(branch) != int:
            raise ValueError('Branch must be an integer')

        # Checking local repository
        lr = self.localRepository.get(plugin, False)
        if lr:
            info = lr.get(str(branch), None)
            if not info:
                versions = [v for v in range(branch, 18, -1)]
                for v in versions:
                    info = lr.get('>=%d' % v, None)
                    if info:
                        break
            if info and info.get('downloadurl'):
                logging.info('Found a compatible version for the plugin in local repository')
                info['component'] = plugin
                info['branch'] = branch
                return PluginDownloadInfo(info)

        # Contacting the remote repository
        data = {
            "branch": round(float(branch) / 10., 1),
            "plugin": plugin
        }

        logging.info('Retrieving information for plugin %s and branch %s' % (data['plugin'], data['branch']))
        try:
            resp = self.request('pluginfo.php', 'GET', data)
        except PluginRepositoryNotFoundException:
            logging.info('No result found')
            return False
        except PluginRepositoryException:
            logging.warning('Error while retrieving information from the plugin database')
            return False

        pluginfo = resp.get('data', {}).get('pluginfo', {})
        pluginfo['branch'] = branch

        return PluginDownloadInfo(pluginfo)

    def request(self, uri, method, data, headers={}):
        """Sends a request to the server and returns the response status and data"""

        uri = self.uri + '/' + str(self.apiversion) + '/' + uri.strip('/')
        method = method.upper()
        if method == 'GET':
            if type(data) == dict:
                data = urlencode(data)
            uri += '?%s' % (data)
            data = ''

        if self.ssl:
            r = httplib.HTTPSConnection(self.host)
        else:
            r = httplib.HTTPConnection(self.host)
        logging.debug('%s %s%s' % (method, self.host, uri))
        r.request(method, uri, data, headers)

        resp = r.getresponse()
        if resp.status == 404:
            raise PluginRepositoryNotFoundException()
        elif resp.status != 200:
            raise PluginRepositoryException('Error during the request to the plugin database')

        data = resp.read()
        if len(data) > 0:
            try:
                data = json.loads(data)
            except ValueError:
                raise PluginRepositoryException('Could not parse JSON data. Data received:\n%s' % data)

        return {'status': resp.status, 'data': data}


class PluginRepositoryException(Exception):
    pass


class PluginRepositoryNotFoundException(Exception):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2012 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import json
from .tools import question
from .config import Conf
from urlparse import urlparse
from datetime import datetime
import re
import logging
import os
import requests
import mimetypes
try:
    import keyring
except:
    logging.warning('Could not load module keyring. You might want to install it.')
    logging.warning('Try `apt-get install python-keyring`, or visit https://pypi.python.org/pypi/keyring#installation-instructions')

C = Conf()


class Jira(object):

    username = ''
    password = ''
    apiversion = '2'
    version = None
    url = None

    host = ''
    ssl = False
    uri = ''

    _loaded = False
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Jira, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        self.version = {}
        self._load()


    def addComment(self, key, comment):
        """Add a comment to an issue"""

        assert isinstance(comment, str), "Expected a string"

        data = [
            {'add': {'body': comment}}
        ]

        self.updateIssue(key, {'update': {'comment': data}})

        return True

    def addLabels(self, key, labels):
        """Add labels to an issue"""

        assert isinstance(labels, list), "Expected a list of labels"

        data = []
        for label in labels:
            data.append({'add': label})

        self.updateIssue(key, {'update': {'labels': data}})

        return True

    def deleteAttachment(self, attachmentId):
        """Deletes an attachment"""
        resp = self.request('attachment/%s' % str(attachmentId), method='DELETE')
        if resp['status'] != 204:
            raise JiraException('Could not delete the attachment')
        return True

    def download(self, url, dest):
        """Download a URL to the destination while authenticating the user"""

        r = requests.get(url, auth=requests.auth.HTTPBasicAuth(self.username, self.password))
        if r.status_code == 403:
            raise JiraException('403 Request not authorized. %s %s' % ('GET', url))

        data = r.text
        if len(data) > 0:
            f = open(dest, 'w')
            f.write(data.encode('utf8'))
            f.close()

        return os.path.isfile(dest)

    def get(self, param, default=None):
        """Returns a property of this instance"""
        return self.info().get(param, default)

    def getAttachments(self, key):
        """Get a dict of attachments

            The keys are the filenames, the values are another dict containing:
            - id: The file ID on the Tracker
            - filename: The file name
            - URL: The URL to download the file
            - date: A datetime object representing the date at which the file was created
            - mimetype: The mimetype of the file
            - size: The size of the file in bytes
            - author: The username of the author of the file
        """
        issueInfo = self.getIssue(key, fields='attachment')
        results = issueInfo.get('fields').get('attachment', [])
        attachments = {}
        for attachment in results:
            attachments[attachment.get('filename')] = {
                'id': attachment.get('id'),
                'filename': attachment.get('filename'),
                'url': attachment.get('content'),
                'date': Jira.parseDate(attachment.get('created')),
                'mimetype': attachment.get('mimeType'),
                'size': attachment.get('size'),
                'author': attachment.get('author', {}).get('name'),
            }
        return attachments

    def getIssue(self, key, fields='*all,-comment'):
        """Load the issue info from the jira server using a rest api call.

        The returned key 'named' of the returned dict is organised by name of the fields, not id.
        """

        querystring = {'fields': fields, 'expand': 'names'}
        resp = self.request('issue/%s' % (str(key)), params=querystring)

        if resp['status'] == 404:
            raise JiraIssueNotFoundException('Issue could not be found.')
        elif not resp['status'] == 200:
            raise JiraException('The tracker is not available.')

        issue = resp['data']
        issue['named'] = {}

        # Populate the named fields in a separate key. Allows us to easily find them without knowing the field ID.
        namelist = issue.get('names', {})
        for fieldkey, fieldvalue in issue.get('fields', {}).items():
            if namelist.get(fieldkey, None) != None:
                issue['named'][namelist.get(fieldkey)] = fieldvalue

        return issue

    def getPullInfo(self, key):
        """Get the pull information organised by branch"""

        fields = self.getIssue(key).get('named')
        infos = {
            'repo': None,
            'branches': {}
        }

        for key, value in C.get('tracker.fieldnames').iteritems():
            if key == 'repositoryurl':
                infos['repo'] = fields.get(value)

            elif key == 'master' or key.isdigit():
                infos['branches'][key] = {
                    'branch': fields.get(value['branch']),
                    'compare': fields.get(value['diffurl'])
                }
            else:
                # We don't know that field...
                continue

        return infos

    def getServerInfo(self):
        """Load the version info from the jira server using a rest api call"""
        resp = self.request('serverInfo')
        if resp['status'] != 200:
            raise JiraException('Unexpected response code: %s' % (str(resp['status'])))

        self.version = resp['data']

    def info(self):
        """Returns a dictionary of information about this instance"""
        info = {}
        self._load()
        for (k, v) in self.version.items():
            info[k] = v
        return info

    def isSecurityIssue(self, key):
        """Return whether or not the issue could be a security issue"""
        resp = self.getIssue(key, fields='security')
        return True if resp.get('fields', {}).get('security', None) != None else False

    def _load(self):
        """Loads the information"""

        if self._loaded:
            return True

        # First get the jira details from the config file.
        self.url = C.get('tracker.url').rstrip('/')
        self.username = C.get('tracker.username')

        parsed = urlparse(self.url)
        self.ssl = True if parsed.scheme == 'https' else False
        self.host = parsed.netloc
        self.uri = parsed.path

        try:
            # str() is needed because keyring does not handle unicode.
            self.password = keyring.get_password('mdk-jira-password', str(self.username))
        except:
            # Do not die if keyring package is not available.
            self.password = None

        if not self.url:
            raise JiraException('The tracker host has not been configured in the config file.')

        askUsername = True if not self.username else False
        while not self._loaded:

            # Testing basic auth
            if self.password:
                try:
                    self.getServerInfo()
                    self._loaded = True
                except JiraException:
                    askUsername = True
                    print 'Either the username and password don\'t match or you may need to enter a Captcha to continue.'
            if not self._loaded:
                if askUsername:
                    self.username = question('What is the username to use to connect to Moodle Tracker?', default=self.username if self.username else None)
                self.password = question('Enter the password for username \'%s\' on Moodle Tracker?' % self.username, password=True)

        # Save the username to the config file
        if self.username != C.get('tracker.username'):
            C.set('tracker.username', self.username)

        try:
            keyring.set_password('mdk-jira-password', str(self.username), str(self.password))
        except:
            # Do not die if keyring package is not available.
            pass

        return True

    def reload(self):
        """Reloads the information"""
        self._loaded = False
        return self._load()

    def request(self, uri, method='GET', data='', params={}, headers={}, files=None):
        """Sends a request to the server and returns the response status and data"""

        url = self.url + self.uri + '/rest/api/' + str(self.apiversion) + '/' + uri.strip('/')

        # Define method to method to use.
        method = method.upper()
        if method == 'GET':
            call = requests.get
        elif method == 'POST':
            call = requests.post
        elif method == 'PUT':
            call = requests.put
        elif method == 'DELETE':
            call = requests.delete
        else:
            raise JiraException('Unimplemented method')

        # Headers.
        if not files:
            headers['Content-Type'] = 'application/json'

        # Call.
        r = call(url, params=params, data=data, auth=requests.auth.HTTPBasicAuth(self.username, self.password),
                headers=headers, files=files)
        if r.status_code == 403:
            raise JiraException('403 Request not authorized. %s %s' % (method, uri))

        try:
            data = r.json()
        except:
            data = r.text

        return {'status': r.status_code, 'data': data}

    @staticmethod
    def parseDate(value):
        """Parse a date returned by Jira API and returns a datetime object."""
        # Strips the timezone information because of some issues with %z.
        value = re.sub(r'[+-]\d+$', '', value)
        return datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')

    def removeLabels(self, key, labels):
        """Remove labels from an issue"""

        assert isinstance(labels, list), "Expected a list of labels"

        data = []
        for label in labels:
            data.append({'remove': label})

        self.updateIssue(key, {'update': {'labels': data}})

    def search(self, query):
        return self.request('search', params={'jql': query, 'fields': 'id'})

    def setCustomFields(self, key, updates):
        """Set a list of fields for this issue in Jira

        The updates parameter is a dictionary of key values where the key is the custom field name
        and the value is the new value to set.

        /!\ This only works for fields of type text.
        """

        issue = self.getIssue(key)
        update = {'fields': {}}

        for updatename, updatevalue in updates.items():
            remotevalue = issue.get('named').get(updatename)
            if not remotevalue or remotevalue != updatevalue:
                # Map the label of the field with the field code.
                fieldKey = None
                for k, v in issue.get('names').iteritems():
                    if v == updatename:
                        fieldKey = k
                        break
                if not fieldKey:
                    raise JiraException('Could not find the field named \'%s\'' % (updatename))
                update['fields'][fieldKey] = updatevalue

        if not update['fields']:
            # No fields to update.
            logging.info('No updates required')
            return True

        resp = self.request('issue/%s' % (str(key)), method='PUT', data=json.dumps(update))

        if resp['status'] != 204:
            raise JiraException('Issue was not updated: %s' % (str(resp['status'])))

        return True

    def updateIssue(self, key, data):
        """Update an issue, the data must be well formatted"""
        resp = self.request('issue/%s' % (str(key)), method='PUT', data=json.dumps(data))

        if resp['status'] != 204:
            raise JiraException('Could not update the issue')

        return True

    def upload(self, key, filepath):
        """Uploads a new attachment to the issue"""

        uri = 'issue/' + key + '/attachments'

        mimetype = mimetypes.guess_type(filepath)[0]
        if not mimetype:
            mimetype = 'application/octet-stream'

        files = {
            'file': (os.path.basename(filepath), open(filepath, 'rb'), mimetype)
        }

        headers = {
            'X-Atlassian-Token': 'nocheck'
        }

        resp = self.request(uri, method='POST', files=files, headers=headers)
        if resp.get('status') != 200:
            raise JiraException('Could not upload file to Jira issue')

        return True


class JiraException(Exception):
    pass


class JiraIssueNotFoundException(JiraException):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
import re
from ..command import Command
from ..tools import launchEditor, yesOrNo
from ..config import ConfigObject, ConfigFileCouldNotBeLoaded


class ConfigCommand(Command):

    _arguments = [
        (
            ['action'],
            {
                'help': 'the action to perform',
                'metavar': 'action',
                'sub-commands': {
                    'edit': (
                        {
                            'help': 'opens the config file in an editor'
                        },
                        []
                    ),
                    'flatlist': (
                        {
                            'help': 'flat list of the settings'
                        },
                        [
                            (
                                ['-g', '--grep'],
                                {
                                    'metavar': 'string',
                                    'help': 'filter results using regular expressions'
                                }
                            ),
                        ]
                    ),
                    'list': (
                        {
                            'help': 'list the settings'
                        },
                        []
                    ),
                    'show': (
                        {
                            'help': 'display one setting'
                        },
                        [
                            (
                                ['setting'],
                                {
                                    'metavar': 'setting',
                                    'help': 'setting to display'
                                }
                            )
                        ]
                    ),
                    'set': (
                        {
                            'help': 'set the value of a setting'
                        },
                        [
                            (
                                ['-b', '--bool'],
                                {
                                    'action': 'store_const',
                                    'dest': 'type',
                                    'const': 'bool',
                                    'help': 'treat the value passed as a boolean'
                                }
                            ),
                            (
                                ['-i', '--int'],
                                {
                                    'action': 'store_const',
                                    'dest': 'type',
                                    'const': 'integer',
                                    'help': 'treat the value passed as an integer'
                                }
                            ),
                            (
                                ['-s', '--string'],
                                {
                                    'action': 'store_const',
                                    'dest': 'type',
                                    'const': 'string',
                                    'help': 'treat the value passed as a string'
                                }
                            ),
                            (
                                ['setting'],
                                {
                                    'metavar': 'setting',
                                    'help': 'setting to edit'
                                }
                            ),
                            (
                                ['value'],
                                {
                                    'default': '',
                                    'metavar': 'value',
                                    'nargs': '?',
                                    'help': 'value to set'
                                }
                            )
                        ]
                    )
                }
            }
        )
    ]
    _description = 'Manage your configuration'

    def dictDisplay(self, data, ident=0):
        for name in sorted(data.keys()):
            setting = data[name]
            if type(setting) != dict:
                print u'{0:<20}: {1}'.format(u' ' * ident + name, setting)
            else:
                print u' ' * ident + '[%s]' % name
                self.dictDisplay(setting, ident + 2)

    def flatDisplay(self, data, parent='', regex=None):
        for name in sorted(data.keys()):
            setting = data[name]
            if type(setting) != dict:
                if regex and not regex.search(parent + name):
                    continue
                print u'%s: %s' % (parent + name, setting)
            else:
                self.flatDisplay(setting, parent=parent + name + u'.', regex=regex)

    def run(self, args):
        if args.action == 'list':
            self.dictDisplay(self.C.get(), 0)

        elif args.action == 'edit':
            f = self.C.userFile
            success = None

            while True:
                tmpfile = launchEditor(filepath=f, suffix='.json')
                co = ConfigObject()
                try:
                    co.loadFromFile(tmpfile)
                except ConfigFileCouldNotBeLoaded:
                    success = False
                    logging.error('I could not load the file, you probably screwed up the JSON...')
                    if yesOrNo('Would you like to continue editing? If not the changes will be discarded.'):
                        f = tmpfile
                        continue
                    else:
                        break
                else:
                    success = True
                    break

            if success:
                with open(tmpfile, 'r') as new:
                    with open(self.C.userFile, 'w') as current:
                        current.write(new.read())

        elif args.action == 'flatlist':
            regex = None
            if args.grep:
                regex = re.compile(args.grep, re.I)
            self.flatDisplay(self.C.get(), regex=regex)

        elif args.action == 'show':
            setting = self.C.get(args.setting)
            if setting != None:
                if type(setting) == dict:
                    self.flatDisplay(setting, args.setting + u'.')
                else:
                    print setting

        elif args.action == 'set':
            type_ = args.type
            setting = args.setting
            val = args.value

            # Classic conversion.
            if type_ == 'bool':
                val = True if val.lower() in ['1', 'true'] else False
            elif type_ == 'integer':
                val = int(val)
            elif type_ == 'string':
                pass

            # Legacy parsing when type isn't stritly defined.
            elif type_ == None and val.startswith('b:'):
                val = True if val[2:].lower() in ['1', 'true'] else False
            elif type_ == None and val.startswith('i:'):
                val = int(val[2:])

            # Catch the rest and try to be smart about the typing.
            else:
                if val in ('1', '0', 'false', 'true'):
                    val = True if val.lower() in ['1', 'true'] else False
                    logging.info('The value was converted to the boolean \'%s\', use the flags --string or '
                        '--int to override this behaviour.', val)

                elif val in ('null', 'none'):
                    val = None
                    logging.info('The value was converted to \'%s\', use the flag --string '
                        'to override this behaviour.', val)

                elif val.isdigit():
                    try:
                        val = int(val)
                        logging.info('The value was converted to an integer, use the flags --string or '
                            '--bool to override this behaviour.')
                    except ValueError:
                        pass

            self.C.set(setting, val)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import re
import os
import logging
from datetime import datetime
from .. import tools, jira, fetch
from ..command import Command
from ..tools import question


class PullCommand(Command):

    _arguments = [
        (
            ['-i', '--integration'],
            {
                'action': 'store_true',
                'help': 'checkout the stable branch before proceeding to the pull. Short for --mode integration.'
            }
        ),
        (
            ['-n', '--no-merge'],
            {
                'action': 'store_true',
                'dest': 'nomerge',
                'help': 'checkout the remote branch without merging. Short for --mode checkout.'
            }
        ),
        (
            ['--fetch-only'],
            {
                'action': 'store_true',
                'dest': 'fetchonly',
                'help': 'only fetches the remote branch, you can then use FETCH_HEAD. Short for --mode fetch.'
            }
        ),
        (
            ['-t', '--testing'],
            {
                'action': 'store_true',
                'help': 'checkout a testing branch before proceeding to the pull. Short for --mode testing.'
            }
        ),
        (
            ['-m', '--mode'],
            {
                'action': 'store',
                'choices': ['checkout', 'fetch', 'integration', 'pull', 'testing'],
                'default': 'pull',
                'help': 'define the mode to use'
            }
        ),
        (
            ['-p', '--prompt'],
            {
                'action': 'store_true',
                'help': 'prompts the user to choose the patch to download.'
            }
        ),
        (
            ['issue'],
            {
                'default': None,
                'help': 'tracker issue to pull from (MDL-12345, 12345). If not specified, read from current branch.',
                'metavar': 'issue',
                'nargs': '?'
            }
        )
    ]
    _description = 'Pull a branch from a tracker issue'

    def run(self, args):

        M = self.Wp.resolve()
        if not M:
            raise Exception('This is not a Moodle instance')

        # Get the mode.
        mode = args.mode
        if args.fetchonly:
            mode = 'fetch'
        elif args.nomerge:
            mode = 'checkout'
        elif args.testing:
            mode = 'testing'
        elif args.integration:
            mode = 'integration'

        # Prompt?
        prompt = args.prompt

        # Tracker issue number.
        issuenb = args.issue
        if not issuenb:
            parsedbranch = tools.parseBranch(M.currentBranch())
            if not parsedbranch:
                raise Exception('Could not extract issue number from %s' % M.currentBranch())
            issuenb = parsedbranch['issue']

        issue = re.sub(r'(MDL|mdl)(-|_)?', '', issuenb)
        mdl = 'MDL-' + issue

        # Reading the information about the current instance.
        branch = M.get('branch')

        # Get information from Tracker
        logging.info('Retrieving information about %s from Moodle Tracker' % (mdl))
        fetcher = fetch.FetchTracker(M)

        try:
            if not prompt:
                fetcher.setFromTracker(mdl, branch)
        except (fetch.FetchTrackerRepoException, fetch.FetchTrackerBranchException) as e:
            prompt = True

        if prompt:
            patches = self.pickPatches(mdl)
            if not patches:
                raise Exception('Could not find any relevant information for a successful pull')
            fetcher.usePatches(patches)

        if mode == 'pull':
            fetcher.pull()
        elif mode == 'checkout':
            fetcher.checkout()
        elif mode == 'fetch':
            fetcher.fetch()
        elif mode == 'integration':
            fetcher.pull(into=M.get('stablebranch'))
        elif mode == 'testing':
            i = 0
            while True:
                i += 1
                suffix = 'test' if i <= 1 else 'test' + str(i)
                newBranch = M.generateBranchName(issue, suffix=suffix, version=branch)
                if not M.git().hasBranch(newBranch):
                    break
            fetcher.pull(into=newBranch, track=M.get('stablebranch'))

    def pickPatches(self, mdl):
        """Prompts the user to pick a patch"""

        J = jira.Jira()
        patches = J.getAttachments(mdl)
        patches = {k: v for k, v in patches.items() if v.get('filename').endswith('.patch')}
        toApply = []

        if len(patches) < 1:
            return False

        mapping = {}
        i = 1
        for key in sorted(patches.keys()):
            patch = patches[key]
            mapping[i] = patch
            print '{0:<2}: {1:<60} {2}'.format(i, key[:60], datetime.strftime(patch.get('date'), '%Y-%m-%d %H:%M'))
            i += 1

        while True:
            try:
                ids = question('What patches would you like to apply?')
                if ids.lower() == 'ankit':
                    logging.warning('Sorry, I am unable to punch a child at the moment...')
                    continue
                elif ids:
                    ids = re.split(r'\s*[, ]\s*', ids)
                    toApply = [mapping[int(i)] for i in ids if int(i) in mapping.keys()]
            except ValueError:
                logging.warning('Error while parsing the list of patches, try a little harder.')
                continue
            break

        return toApply




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

def main():

    import sys
    import argparse
    import os
    import re
    import logging
    from .command import CommandRunner
    from .commands import getCommand, commandsList
    from .config import Conf
    from .tools import process
    from .version import __version__

    C = Conf()

    try:
        debuglevel = getattr(logging, C.get('debug').upper())
    except AttributeError:
        debuglevel = logging.INFO

    # Set logging levels.
    logging.basicConfig(format='%(message)s', level=debuglevel)
    logging.getLogger('requests').setLevel(logging.WARNING)  # Reset logging level of 'requests' module.
    logging.getLogger('keyring.backend').setLevel(logging.WARNING)

    availaliases = [str(x) for x in C.get('aliases').keys()]
    choices = sorted(commandsList + availaliases)

    parser = argparse.ArgumentParser(description='Moodle Development Kit', add_help=False)
    parser.add_argument('-h', '--help', action='store_true', help='show this help message and exit')
    parser.add_argument('-l', '--list', action='store_true', help='list the available commands')
    parser.add_argument('-v', '--version', action='store_true', help='display the current version')
    parser.add_argument(*['--%s'%f.decode('base64') for f in ('aWNhbnRyZWFj', 'aWNhbnRyZWFk')], dest='asdf', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('command', metavar='command', nargs='?', help='command to call', choices=choices)
    parser.add_argument('args', metavar='arguments', nargs=argparse.REMAINDER, help='arguments of the command')
    parsedargs = parser.parse_args()

    cmd = parsedargs.command
    args = parsedargs.args

    # There is no command, what do we do?
    if not cmd:
        if parsedargs.version:
            print 'MDK version %s' % __version__
        elif parsedargs.asdf:
            print 'U29ycnkgRGF2ZSwgTURLIGNhbm5vdCBoZWxwIHlvdSB3aXRoIHRoYXQuLi4='.decode('base64')
        elif parsedargs.list:
            for c in sorted(commandsList):
                print '{0:<15} {1}'.format(c, getCommand(c)._description)
        else:
            parser.print_help()
        sys.exit(0)

    # Looking up for an alias
    alias = C.get('aliases.%s' % cmd)
    if alias != None:
        if alias.startswith('!'):
            cmd = alias[1:]
            i = 0
            # Replace $1, $2, ... with passed arguments
            for arg in args:
                i += 1
                cmd = cmd.replace('$%d' % i, arg)
            # Remove unknown $[0-9]
            cmd = re.sub(r'\$[0-9]', '', cmd)
            result = process(cmd, stdout=None, stderr=None)
            sys.exit(result[0])
        else:
            cmd = alias.split(' ')[0]
            args = alias.split(' ')[1:] + args

    cls = getCommand(cmd)
    Cmd = cls(C)
    Runner = CommandRunner(Cmd)
    try:
        Runner.run(args, prog='%s %s' % (os.path.basename(sys.argv[0]), cmd))
    except Exception as e:
        import traceback
        info = sys.exc_info()
        logging.error('%s: %s', e.__class__.__name__, e)
        logging.debug(''.join(traceback.format_tb(info[2])))
        sys.exit(1)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2014 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
import os
from .tools import process
from .config import Conf
from .plugins import PluginManager

C = Conf()


class Js(object):
    """Class wrapping JS related functions"""

    _M = None

    def __init__(self, M):
        self._M = M

    def shift(self, subsystemOrPlugin=None, module=None):
        """Runs shifter"""
        path = self.getYUISrcPath(subsystemOrPlugin, module=module)
        if not os.path.isdir(path):
            raise ValueError("The directory '%s' was not found" % (path))

        paths = []
        if module:
            paths.append(path)

        else:
            dirs = os.listdir(path)
            for d in dirs:
                if os.path.isdir(os.path.join(path, d, 'js')):
                    paths.append(os.path.join(path, d))

        shifter = Shifter(path)
        for path in paths:
            readablePath = path.replace(self._M.get('path'), '')
            logging.info('Shifting in %s' % readablePath)
            shifter.setCwd(path)
            shifter.compile()

    def document(self, outdir):
        """Runs documentator"""

        # TODO We should be able to generate outdir from here, using the workplace.
        path = self._M.get('path')
        documentor = Documentor(path, outdir)
        documentor.compile()

    def getYUISrcPath(self, subsystemOrPlugin, module=None):
        """Returns the path to the module, or the component"""

        try:
            path = PluginManager.getSubsystemDirectory(subsystemOrPlugin, M=self._M)
        except ValueError:
            (pluginType, name) = PluginManager.getTypeAndName(subsystemOrPlugin)
            path = PluginManager.getTypeDirectory(pluginType, M=self._M)
            # An exception will be thrown here if we do not find the plugin or component, that is fine.
            path = os.path.join(path, name)

        path = os.path.join(path, 'yui', 'src')
        if module:
            path = os.path.join(path, module)

        return path


class Shifter(object):

    _cwd = None

    def __init__(self, cwd=None):
        self.setCwd(cwd)

    def compile(self):
        """Runs the shifter command in cwd"""
        executable = C.get('shifter')
        if not executable or not os.path.isfile(executable):
            raise Exception('Could not find executable path %s' % (executable))

        cmd = [executable]
        (code, out, err) = process(cmd, cwd=self._cwd)
        if code != 0:
            raise ShifterCompileFailed('Error during shifting at %s' % (self._cwd))

    def setCwd(self, cwd):
        self._cwd = cwd


class Documentor(object):

    _cwd = None
    _outdir = None

    def __init__(self, cwd=None, outdir=None):
        self.setCwd(cwd)
        self.setOutdir(outdir)

    def compile(self):
        """Runs the yuidoc command in cwd"""
        executable = C.get('yuidoc')
        if not executable or not os.path.isfile(executable):
            raise Exception('Could not find executable path %s' % (executable))

        cmd = [executable, '--outdir', self._outdir]

        (code, out, err) = process(cmd, cwd=self._cwd)
        if code != 0:
            raise YuidocCompileFailed('Error whilst generating documentation')

    def setCwd(self, cwd):
        self._cwd = cwd

    def setOutdir(self, outdir):
        self._outdir = outdir


class YuidocCompileFailed(Exception):
    pass
class ShifterCompileFailed(Exception):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2014 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
import os
import time
import watchdog.events
import watchdog.observers
from .. import css
from ..command import Command


class CssCommand(Command):

    _arguments = [
        (
            ['-c', '--compile'],
            {
                'action': 'store_true',
                'dest': 'compile',
                'help': 'compile the theme less files'
            }
        ),
        (
            ['-s', '--sheets'],
            {
                'action': 'store',
                'dest': 'sheets',
                'default': None,
                'help': 'the sheets to work on without their extensions. When not specified, it is guessed from the less folder.',
                'nargs': '+'
            }
        ),
        (
            ['-t', '--theme'],
            {
                'action': 'store',
                'dest': 'theme',
                'default': None,
                'help': 'the theme to work on. The default is \'bootstrapbase\' but is ignored if we are in a theme folder.',
            }
        ),
        (
            ['-d', '--debug'],
            {
                'action': 'store_true',
                'dest': 'debug',
                'help': 'produce an unminified debugging version with source maps'
            }
        ),
        (
            ['-w', '--watch'],
            {
                'action': 'store_true',
                'dest': 'watch',
                'help': 'watch the directory'
            }
        ),
        (
            ['names'],
            {
                'default': None,
                'help': 'name of the instances',
                'metavar': 'names',
                'nargs': '*'
            }
        )
    ]
    _description = 'Wrapper for CSS functions'

    def run(self, args):

        Mlist = self.Wp.resolveMultiple(args.names)
        if len(Mlist) < 1:
            raise Exception('No instances to work on. Exiting...')

        # Resolve the theme folder we are in.
        if not args.theme:
            mpath = os.path.join(Mlist[0].get('path'), 'theme')
            cwd = os.path.realpath(os.path.abspath(os.getcwd()))
            if cwd.startswith(mpath):
                candidate = cwd.replace(mpath, '').strip('/')
                while True:
                    (head, tail) = os.path.split(candidate)
                    if not head and tail:
                        # Found the theme.
                        args.theme = tail
                        logging.info('You are in the theme \'%s\', using that.' % (args.theme))
                        break
                    elif not head and not tail:
                        # Nothing, let's leave.
                        break
                    candidate = head

        # We have not found anything, falling back on the default.
        if not args.theme:
            args.theme = 'bootstrapbase'

        for M in Mlist:
            if args.compile:
                logging.info('Compiling theme \'%s\' on %s' % (args.theme, M.get('identifier')))
                processor = css.Css(M)
                processor.setDebug(args.debug)
                if args.debug:
                    processor.setCompiler('lessc')
                elif M.branch_compare(29, '<'):
                    # Grunt was only introduced for 2.9.
                    processor.setCompiler('recess')

                processor.compile(theme=args.theme, sheets=args.sheets)

        # Setting up watchdog. This code should be improved when we will have more than a compile option.
        observer = None
        if args.compile and args.watch:
            observer = watchdog.observers.Observer()

        for M in Mlist:
            if args.watch and args.compile:
                processor = css.Css(M)
                processorArgs = {'theme': args.theme, 'sheets': args.sheets}
                handler = LessWatcher(M, processor, processorArgs)
                observer.schedule(handler, processor.getThemeLessPath(args.theme), recursive=True)
                logging.info('Watchdog set up on %s/%s, waiting for changes...' % (M.get('identifier'), args.theme))

        if observer and args.compile and args.watch:
            observer.start()

            try:
                while True:
                    time.sleep(1)
            except KeyboardInterrupt:
                observer.stop()
            finally:
                observer.join()


class LessWatcher(watchdog.events.FileSystemEventHandler):

    _processor = None
    _args = None
    _ext = '.less'
    _M = None

    def __init__(self, M, processor, args):
        super(self.__class__, self).__init__()
        self._M = M
        self._processor = processor
        self._args = args

    def on_modified(self, event):
        self.process(event)

    def on_moved(self, event):
        self.process(event)

    def process(self, event):
        if event.is_directory:
            return
        elif not event.src_path.endswith(self._ext):
            return

        filename = event.src_path.replace(self._processor.getThemeLessPath(self._args['theme']), '').strip('/')
        logging.info('[%s] Changes detected in %s!' % (self._M.get('identifier'), filename))
        self._processor.compile(**self._args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import urllib
import re
import logging
import gzip
from tempfile import gettempdir
from time import sleep
from ..command import Command
from ..tools import process, ProcessInThread, downloadProcessHook, question


class BehatCommand(Command):

    _arguments = [
        (
            ['-r', '--run'],
            {
                'action': 'store_true',
                'help': 'run the tests'
            }
        ),
        (
            ['-d', '--disable'],
            {
                'action': 'store_true',
                'help': 'disable Behat, runs the tests first if --run has been set. Ignored from 2.7.'
            }
        ),
        (
            ['--force'],
            {
                'action': 'store_true',
                'help': 'force behat re-init and reset the variables in the config file.'
            }
        ),
        (
            ['-f', '--feature'],
            {
                'metavar': 'path',
                'help': 'typically a path to a feature, or an argument understood by behat (see [features]: vendor/bin/behat --help). Automatically convert path to absolute path.'
            }
        ),
        (
            ['-n', '--testname'],
            {
                'dest': 'testname',
                'metavar': 'name',
                'help': 'only execute the feature elements which match part of the given name or regex'
            }
        ),
        (
            ['-t', '--tags'],
            {
                'metavar': 'tags',
                'help': 'only execute the features or scenarios with tags matching tag filter expression'
            }
        ),
        (
            ['-j', '--no-javascript'],
            {
                'action': 'store_true',
                'dest': 'nojavascript',
                'help': 'do not start Selenium and ignore Javascript (short for --tags=~@javascript). Cannot be combined with --tags or --testname.'
            }
        ),
        (
            ['-D', '--no-dump'],
            {
                'action': 'store_false',
                'dest': 'faildump',
                'help': 'use the standard command without fancy screenshots or output to a directory'
            }
        ),
        (
            ['-s', '--switch-completely'],
            {
                'action': 'store_true',
                'dest': 'switchcompletely',
                'help': 'force the switch completely setting. This will be automatically enabled for PHP < 5.4. Ignored from 2.7.'
            }
        ),
        (
            ['--selenium'],
            {
                'default': None,
                'dest': 'selenium',
                'help': 'path to the selenium standalone server to use',
                'metavar': 'jarfile'
            }
        ),
        (
            ['--selenium-download'],
            {
                'action': 'store_true',
                'dest': 'seleniumforcedl',
                'help': 'force the download of the latest Selenium to the cache'
            }
        ),
        (
            ['--selenium-verbose'],
            {
                'action': 'store_true',
                'dest': 'seleniumverbose',
                'help': 'outputs the output from selenium in the same window'
            }
        ),
        (
            ['name'],
            {
                'default': None,
                'help': 'name of the instance',
                'metavar': 'name',
                'nargs': '?'
            }
        )
    ]
    _description = 'Initialise Behat'

    def run(self, args):

        # Loading instance
        M = self.Wp.resolve(args.name)
        if not M:
            raise Exception('This is not a Moodle instance')

        # Check required version
        if M.branch_compare(25, '<'):
            raise Exception('Behat is only available from Moodle 2.5')

        # Check if installed
        if not M.get('installed'):
            raise Exception('This instance needs to be installed first')

        # Disable Behat
        if args.disable and not args.run:
            self.disable(M)
            return

        # No Javascript
        nojavascript = args.nojavascript
        if not nojavascript and not self.C.get('java') or not os.path.isfile(os.path.abspath(self.C.get('java'))):
            nojavascript = True
            logging.info('Disabling Javascript because Java is required to run Selenium and could not be found.')

        # If not composer.phar, install Composer
        if not os.path.isfile(os.path.join(M.get('path'), 'composer.phar')):
            logging.info('Installing Composer')
            cliFile = 'behat_install_composer.php'
            cliPath = os.path.join(M.get('path'), 'behat_install_composer.php')
            (to, headers) = urllib.urlretrieve('http://getcomposer.org/installer', cliPath)
            if headers.dict.get('content-encoding') == 'gzip':
                f = gzip.open(cliPath, 'r')
                content = f.read()
                f.close()
                f = open(cliPath, 'w')
                f.write(content)
                f.close()
            M.cli('/' + cliFile, stdout=None, stderr=None)
            os.remove(cliPath)
            M.cli('composer.phar', args='install --dev', stdout=None, stderr=None)

        # Download selenium
        seleniumPath = os.path.expanduser(os.path.join(self.C.get('dirs.mdk'), 'selenium.jar'))
        if args.selenium:
            seleniumPath = args.selenium
        elif args.seleniumforcedl or (not nojavascript and not os.path.isfile(seleniumPath)):
            logging.info('Attempting to find a download for Selenium')
            seleniumStorageUrl = 'https://selenium-release.storage.googleapis.com/'
            url = urllib.urlopen(seleniumStorageUrl)
            content = url.read()
            matches = sorted(re.findall(r'[a-z0-9/._-]+selenium-server-standalone-[0-9.]+\.jar', content, re.I))
            if len(matches) > 0:
                seleniumUrl = seleniumStorageUrl + matches[-1]
                logging.info('Downloading Selenium from %s' % seleniumUrl)
                if (logging.getLogger().level <= logging.INFO):
                    urllib.urlretrieve(seleniumUrl, seleniumPath, downloadProcessHook)
                    # Force a new line after the hook display
                    logging.info('')
                else:
                    urllib.urlretrieve(seleniumUrl, seleniumPath)
            else:
                logging.warning('Could not locate Selenium server to download')

        if not nojavascript and not os.path.isfile(seleniumPath):
            raise Exception('Selenium file %s does not exist')

        # Run cli
        try:

            # If Oracle, ask the user for a Behat prefix, if not set.
            prefix = M.get('behat_prefix')
            if M.get('dbtype') == 'oci' and (args.force or not prefix or len(prefix) > 2):
                while not prefix or len(prefix) > 2:
                    prefix = question('What prefix would you like to use? (Oracle, max 2 chars)')
            else:
                prefix = None

            outputDir = self.Wp.getExtraDir(M.get('identifier'), 'behat')
            outpurUrl = self.Wp.getUrl(M.get('identifier'), extra='behat')

            logging.info('Initialising Behat, please be patient!')
            M.initBehat(switchcompletely=args.switchcompletely, force=args.force, prefix=prefix, faildumppath=outputDir)
            logging.info('Behat ready!')

            # Preparing Behat command
            cmd = ['vendor/bin/behat']
            if args.tags:
                cmd.append('--tags="%s"' % (args.tags))

            if args.testname:
                cmd.append('--name="%s"' % (args.testname))

            if not (args.tags or args.testname) and nojavascript:
                cmd.append('--tags ~@javascript')

            if args.faildump:
                if M.branch_compare(31, '<'):
                    cmd.append('--format="progress,progress,pretty,html,failed"')
                    cmd.append('--out=",{0}/progress.txt,{0}/pretty.txt,{0}/status.html,{0}/failed.txt"'.format(outputDir))
                else:
                    cmd.append('--format="moodle_progress" --out="std"')
                    cmd.append('--format="progress" --out="{0}/progress.txt"'.format(outputDir))
                    cmd.append('--format="pretty" --out="{0}/pretty.txt"'.format(outputDir))

            configcandidates = ['%s/behat/behat.yml' % (M.get('behat_dataroot'))]
            if M.branch_compare(32):
                # Since Moodle 3.2.2 behat directory is kept under $CFG->behat_dataroot for single and parallel runs.
                configcandidates.insert(0, '%s/behatrun/behat/behat.yml' % (M.get('behat_dataroot')))

            cmd.append('--config=%s' % (filter(os.path.isfile, configcandidates)[0]))

            # Checking feature argument
            if args.feature:
                filepath = args.feature
                if not filepath.startswith('/'):
                    filepath = os.path.join(M.get('path'), filepath)
                cmd.append(filepath)

            cmd = ' '.join(cmd)

            phpCommand = '%s -S localhost:8000' % (self.C.get('php'))
            seleniumCommand = None
            if seleniumPath:
                seleniumCommand = '%s -jar %s' % (self.C.get('java'), seleniumPath)

            olderThan26 = M.branch_compare(26, '<')

            if args.run:
                logging.info('Preparing Behat testing')

                # Preparing PHP Server
                phpServer = None
                if olderThan26 and not M.get('behat_switchcompletely'):
                    logging.info('Starting standalone PHP server')
                    kwargs = {}
                    kwargs['cwd'] = M.get('path')
                    phpServer = ProcessInThread(phpCommand, **kwargs)
                    phpServer.start()

                # Launching Selenium
                seleniumServer = None
                if seleniumPath and not nojavascript:
                    logging.info('Starting Selenium server')
                    kwargs = {}
                    if args.seleniumverbose:
                        kwargs['stdout'] = None
                        kwargs['stderr'] = None
                    else:
                        # Logging Selenium to a temporary file, this can be useful, and also it appears
                        # that Selenium hangs when stderr is not buffered.
                        fileOutPath = os.path.join(gettempdir(), 'selenium_%s_out.log' % (M.get('identifier')))
                        fileErrPath = os.path.join(gettempdir(), 'selenium_%s_err.log' % (M.get('identifier')))
                        tmpfileOut = open(fileOutPath, 'w')
                        tmpfileErr = open(fileErrPath, 'w')
                        logging.debug('Logging Selenium output to: %s' % (fileOutPath))
                        logging.debug('Logging Selenium errors to: %s' % (fileErrPath))
                        kwargs['stdout'] = tmpfileOut
                        kwargs['stderr'] = tmpfileErr
                    seleniumServer = ProcessInThread(seleniumCommand, **kwargs)
                    seleniumServer.start()

                logging.info('Running Behat tests')

                # Sleep for a few seconds before starting Behat
                if phpServer or seleniumServer:
                    launchSleep = int(self.C.get('behat.launchSleep'))
                    logging.debug('Waiting for %d seconds to allow Selenium and/or the PHP Server to start ' % (launchSleep))
                    sleep(launchSleep)

                # Running the tests
                try:
                    if args.faildump:
                        logging.info('More output can be found at:\n %s\n %s', outputDir, outpurUrl)
                    process(cmd, M.path, None, None)
                except KeyboardInterrupt:
                    pass

                # Kill the remaining processes
                if phpServer and phpServer.is_alive():
                    phpServer.kill()
                if seleniumServer and seleniumServer.is_alive():
                    seleniumServer.kill()

                # Disable Behat
                if args.disable:
                    self.disable(M)

            else:
                if args.faildump:
                    logging.info('More output will be accessible at:\n %s\n %s', outputDir, outpurUrl)
                if olderThan26:
                    logging.info('Launch PHP Server (or set $CFG->behat_switchcompletely to True):\n %s' % (phpCommand))
                if seleniumCommand:
                    logging.info('Launch Selenium (optional):\n %s' % (seleniumCommand))
                logging.info('Launch Behat:\n %s' % (cmd))

        except Exception as e:
            raise e

    def disable(self, M):
        logging.info('Disabling Behat')
        M.cli('admin/tool/behat/cli/util.php', '--disable')
        M.removeConfig('behat_switchcompletely')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import shutil
import stat
import logging
from pkg_resources import resource_filename
from .tools import process
from .config import Conf
from .exceptions import ScriptNotFound, ConflictInScriptName, UnsupportedScript

C = Conf()


class Scripts(object):

    _supported = ['php', 'sh']
    _dirs = None
    _list = None

    @classmethod
    def dirs(cls):
        """Return the directories containing scripts, in priority order"""

        if not cls._dirs:
            dirs = ['~/.moodle-sdk']
            if C.get('dirs.moodle') != None:
                dirs.insert(0, C.get('dirs.moodle'))

            dirs.append('/etc/moodle-sdk')

            # Directory within the package.
            # This can point anywhere when the package is installed, or to the folder containing the module when it is not.
            packageDir = resource_filename('mdk', 'scripts')
            dirs.append(os.path.split(packageDir)[0])

            # Legacy: directory part of the root git repository, only if we can be sure that the parent directory is still MDK.
            if os.path.isfile(os.path.join(os.path.dirname(__file__), '..', 'mdk.py')):
                dirs.append(os.path.join(os.path.dirname(__file__), '..'))

            i = 0
            for d in dirs:
                dirs[i] = os.path.expanduser(os.path.join(d, 'scripts'))
                i += 1

            cls._dirs = dirs

        return cls._dirs

    @classmethod
    def list(cls):
        """Return a dict where keys are the name of the scripts
           and the value is the directory in which the script is stored"""

        if not cls._list:
            scripts = {}

            # Walk through the directories, in reverse to get the higher
            # priority last.
            dirs = cls.dirs()
            dirs.reverse()
            for d in dirs:

                if not os.path.isdir(d):
                    continue

                # For each file found in the directory.
                l = os.listdir(d)
                for f in l:

                    # Check if supported format.
                    supported = False
                    for ext in cls._supported:
                        if f.endswith('.' + ext):
                            supported = True
                            break

                    if supported:
                        scripts[f] = d

            cls._list = scripts

        return cls._list

    @classmethod
    def find(cls, script):
        """Return the path to a script"""

        lst = cls.list()
        cli = None
        if script in lst.keys():
            cli = os.path.join(lst[script], script)
        else:
            found = 0
            for ext in cls._supported:
                candidate = script + '.' + ext
                if candidate in lst.keys():
                    scriptFile = candidate
                    found += 1

            if found > 1:
                raise ConflictInScriptName('The script name conflicts with other ones')
            elif found == 1:
                cli = os.path.join(lst[scriptFile], scriptFile)

        if not cli:
            raise ScriptNotFound('Script not found')

        return cli

    @classmethod
    def get_script_destination(cls, cli, path):
        """Get the final path where the script will be copied"""

        ext = os.path.splitext(cli)[1]

        i = 0
        while True:
            candidate = os.path.join(path, 'mdkscriptrun{}{}'.format(i if i > 0 else '', ext))
            if not os.path.isfile(candidate):
                break
            i += 1

        return candidate

    @classmethod
    def run(cls, script, path, arguments=None, cmdkwargs={}):
        """Executes a script at in a certain directory"""

        # Converts arguments to a string.
        arguments = '' if arguments == None else arguments
        if type(arguments) == list:
            arguments = ' '.join(arguments)
        arguments = ' ' + arguments

        cli = cls.find(script)
        dest = cls.get_script_destination(cli, path)

        if cli.endswith('.php'):
            logging.debug('Copying %s to %s' % (cli, dest))
            shutil.copyfile(cli, dest)

            cmd = '%s %s %s' % (C.get('php'), dest, arguments)

            result = process(cmd, cwd=path, **cmdkwargs)
            os.remove(dest)
        elif cli.endswith('.sh'):
            logging.debug('Copying %s to %s' % (cli, dest))
            shutil.copyfile(cli, dest)
            os.chmod(dest, stat.S_IRUSR | stat.S_IXUSR)

            cmd = '%s %s' % (dest, arguments)
            result = process(cmd, cwd=path, **cmdkwargs)
            os.remove(dest)
        else:
            raise UnsupportedScript('Script not supported')

        return result[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2012 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import re
import logging
import shutil
import subprocess
from tempfile import gettempdir

from .tools import getMDLFromCommitMessage, mkdir, process, parseBranch
from .db import DB
from .config import Conf
from .git import Git, GitException
from .exceptions import InstallException, UpgradeNotAllowed
from .jira import Jira, JiraException
from .scripts import Scripts

C = Conf()


class Moodle(object):

    identifier = None
    path = None
    installed = False
    version = None
    config = None

    _dbo = None
    _git = None
    _loaded = False

    _cos_hasstash = False
    _cos_oldbranch = None

    _reservedKeywords = [
        'branch',
        'identifier',
        'installed',
        'integration',
        'maturity',
        'path',
        'release',
        'stablebranch',
        'version'
    ]

    def __init__(self, path, identifier=None):
        self.path = path
        self.identifier = identifier
        self.version = {}
        self.config = {}
        self._load()

    def addConfig(self, name, value):
        """Add a parameter to the config file
        Will attempt to write them before the inclusion of lib/setup.php"""
        configFile = os.path.join(self.path, 'config.php')
        if not os.path.isfile(configFile):
            return None

        if name in self._reservedKeywords:
            raise Exception('Cannot use reserved keywords for settings in config.php')

        if type(value) == bool:
            value = 'true' if value else 'false'
        elif type(value) != int:
            value = "'" + str(value) + "'"
        value = str(value)

        try:
            f = open(configFile, 'r')
            lines = f.readlines()
            f.close()

            for i, line in enumerate(lines):
                if re.search(r'^// MDK Edit\.$', line.rstrip()):
                    break
                elif re.search(r'require_once.*/lib/setup\.php', line):
                    lines.insert(i, '// MDK Edit.\n')
                    lines.insert(i + 1, '\n')
                    # As we've added lines, let's move the index
                    break

            i += 1
            if i > len(lines):
                i = len(lines)
            lines.insert(i, '$CFG->%s = %s;\n' % (name, value))

            f = open(configFile, 'w')
            f.writelines(lines)
            f.close()
        except:
            raise Exception('Error while writing to config file')

        self.reload()

    def branch_compare(self, branch, compare='>='):
        """Compare the branch of the current instance with the one passed"""
        try:
            branch = int(branch)
        except:
            raise Exception('Could not convert branch to int, got %s' % branch)
        b = self.get('branch')
        if b == None:
            raise Exception('Error while reading the branch')
        elif b == 'master':
            b = C.get('masterBranch')
        b = int(b)
        if compare == '>=':
            return b >= branch
        elif compare == '>':
            return b > branch
        elif compare == '=' or compare == '==':
            return b == branch
        if compare == '<=':
            return b <= branch
        elif compare == '<':
            return b < branch
        return False

    def checkout_stable(self, checkout=True):
        """Checkout the stable branch, do a stash if required. Needs to be called again to pop the stash!"""

        # Checkout the branch
        if checkout:
            stablebranch = self.get('stablebranch')
            if self.currentBranch() == stablebranch:
                self._cos_oldbranch = None
                return True

            self._cos_oldbranch = self.currentBranch()
            self._cos_hasstash = False

            # Stash
            stash = self.git().stash(untracked=True)
            if stash[0] != 0:
                raise Exception('Error while stashing your changes')
            if not stash[1].startswith('No local changes'):
                self._cos_hasstash = True

            # Checkout STABLE
            if not self.git().checkout(stablebranch):
                raise Exception('Could not checkout %s' % stablebranch)

        # Checkout the previous branch
        elif self._cos_oldbranch != None:
            if not self.git().checkout(self._cos_oldbranch):
                raise Exception('Could not checkout working branch %s' % self._cos_oldbranch)

            # Unstash
            if self._cos_hasstash:
                pop = self.git().stash('pop')
                if pop[0] != 0:
                    raise Exception('Error while popping the stash. Probably got conflicts.')
                self._cos_hasstash = False

    def cli(self, cli, args='', **kwargs):
        """Executes a command line tool script"""
        cli = os.path.join(self.get('path'), cli.lstrip('/'))
        if not os.path.isfile(cli):
            raise Exception('Could not find script to call')
        if type(args) == 'list':
            args = ' '.join(args)
        cmd = '%s %s %s' % (C.get('php'), cli, args)
        return process(cmd, cwd=self.get('path'), **kwargs)

    def currentBranch(self):
        """Returns the current branch on the git repository"""
        return self.git().currentBranch()

    def dbo(self):
        """Returns a Database object"""
        if self._dbo == None:
            engine = self.get('dbtype')
            db = self.get('dbname')
            if engine != None and db != None:
                try:
                    self._dbo = DB(engine, C.get('db.%s' % engine))
                except:
                    pass
        return self._dbo

    def generateBranchName(self, issue, suffix='', version=''):
        """Generates a branch name"""
        mdl = re.sub(r'(MDL|mdl)(-|_)?', '', issue)
        if version == '':
            version = self.get('branch')
        args = {
            'issue': mdl,
            'version': version
        }
        branch = C.get('wording.branchFormat') % args
        if suffix != None and suffix != '':
            branch += C.get('wording.branchSuffixSeparator') + suffix
        return branch

    def get(self, param, default=None):
        """Returns a property of this instance"""
        info = self.info()
        try:
            return info[param]
        except:
            return default

    def git(self):
        """Returns a Git object"""
        if self._git == None:
            self._git = Git(self.path, C.get('git'))
            if not self._git.isRepository():
                raise Exception('Could not find the Git repository')
        return self._git

    def headcommit(self, branch=None):
        """Try to resolve the head commit of branch of this instance"""

        if branch == None:
            branch = self.currentBranch()
            if branch == 'HEAD':
                raise Exception('Cannot update the tracker when on detached branch')

        smartSearch = C.get('smartHeadCommitSearch')

        # Parsing the branch
        parsedbranch = parseBranch(branch)
        if parsedbranch:
            issue = 'MDL-%s' % (parsedbranch['issue'])
        else:
            logging.debug('Cannot smart resolve using the branch %s' % (branch))
            smartSearch = False

        headcommit = None
        try:
            # Trying to smart guess the last commit needed
            if smartSearch:
                commits = self.git().log(since=branch, count=C.get('smartHeadCommitLimit'), format='%s_____%H').split('\n')[:-1]

                # Looping over the last commits to find the commit messages that match the MDL-12345.
                candidate = None
                for commit in commits:
                    match = getMDLFromCommitMessage(commit) == issue
                    if not candidate and not match:
                        # The first commit does not match a hash, let's ignore this method.
                        break
                    candidate = commit.split('_____')[-1]
                    if not match:
                        # The commit does not match any more, we found it!
                        headcommit = candidate
                        break

            # We could not smart find the last commit, let's use the default mechanism.
            if not headcommit:
                upstreamremote = C.get('upstreamRemote')
                stablebranch = self.get('stablebranch')
                headcommit = self.git().hashes(ref='%s/%s' % (upstreamremote, stablebranch), limit=1, format='%H')[0]

        except GitException:
            logging.warning('Could not resolve the head commit')
            headcommit = False

        return headcommit

    def initPHPUnit(self, force=False, prefix=None):
        """Initialise the PHPUnit environment"""
        raise Exception('This method is deprecated, use phpunit.PHPUnit.init() instead.')

    def initBehat(self, switchcompletely=False, force=False, prefix=None, faildumppath=None):
        """Initialise the Behat environment"""

        if self.branch_compare(25, '<'):
            raise Exception('Behat is only available from Moodle 2.5')

        # Force switch completely for PHP < 5.4
        (none, phpVersion, none) = process('%s -r "echo version_compare(phpversion(), \'5.4\');"' % (C.get('php')))
        if int(phpVersion) <= 0:
            switchcompletely = True

        # Set Behat data root
        behat_dataroot = self.get('dataroot') + '_behat'
        self.updateConfig('behat_dataroot', behat_dataroot)

        # Set Behat DB prefix
        currentPrefix = self.get('behat_prefix')
        behat_prefix = prefix or 'zbehat_'

        # Set behat_faildump_path
        currentFailDumpPath = self.get('behat_faildump_path')
        if faildumppath and currentFailDumpPath != faildumppath:
            self.updateConfig('behat_faildump_path', faildumppath)
        elif (not faildumppath and currentFailDumpPath):
            self.removeConfig('behat_faildump_path')

        if not currentPrefix or force:
            self.updateConfig('behat_prefix', behat_prefix)
        elif currentPrefix != behat_prefix and self.get('dbtype') != 'oci':
            # Warn that a prefix is already set and we did not change it.
            # No warning for Oracle as we need to set it to something else.
            logging.warning('Behat prefix not changed, already set to \'%s\', expected \'%s\'.' % (currentPrefix, behat_prefix))

        # Switch completely?
        if self.branch_compare(26, '<'):
            if switchcompletely:
                self.updateConfig('behat_switchcompletely', switchcompletely)
                self.updateConfig('behat_wwwroot', self.get('wwwroot'))
            else:
                self.removeConfig('behat_switchcompletely')
                self.removeConfig('behat_wwwroot')
        else:
            # Defining wwwroot.
            wwwroot = '%s://%s/' % (C.get('scheme'), C.get('behat.host'))
            if C.get('path') != '' and C.get('path') != None:
                wwwroot = wwwroot + C.get('path') + '/'
            wwwroot = wwwroot + self.identifier
            currentWwwroot = self.get('behat_wwwroot')
            if not currentWwwroot or force:
                self.updateConfig('behat_wwwroot', wwwroot)
            elif currentWwwroot != wwwroot:
                logging.warning('Behat wwwroot not changed, already set to \'%s\', expected \'%s\'.' % (currentWwwroot, wwwroot))

        # Force a cache purge
        self.purge()

        # Force dropping the tables if there are any.
        if force:
            result = self.cli('admin/tool/behat/cli/util.php', args='--drop', stdout=None, stderr=None)
            if result[0] != 0:
                raise Exception('Error while initialising Behat. Please try manually.')

        # Run the init script.
        result = self.cli('admin/tool/behat/cli/init.php', stdout=None, stderr=None)
        if result[0] != 0:
            raise Exception('Error while initialising Behat. Please try manually.')

        # Force a cache purge
        self.purge()

    def info(self):
        """Returns a dictionary of information about this instance"""
        self._load()
        info = {
            'path': self.path,
            'installed': self.isInstalled(),
            'identifier': self.identifier
        }
        for (k, v) in self.config.items():
            info[k] = v
        for (k, v) in self.version.items():
            info[k] = v
        return info

    def install(self, dbname=None, engine=None, dataDir=None, fullname=None, dropDb=False, wwwroot=None):
        """Launch the install script of an Instance"""

        if self.isInstalled():
            raise InstallException('Instance already installed!')

        if not wwwroot:
            raise InstallException('Cannot install without a value for wwwroot')
        if dataDir == None or not os.path.isdir(dataDir):
            raise InstallException('Cannot install instance without knowing where the data directory is')
        if dbname == None:
            dbname = re.sub(r'[^a-zA-Z0-9]', '', self.identifier).lower()
            prefixDbname = C.get('db.namePrefix')
            if prefixDbname:
                dbname = prefixDbname + dbname
            dbname = dbname[:28]
        if engine == None:
            engine = C.get('defaultEngine')
        if fullname == None:
            fullname = self.identifier.replace('-', ' ').replace('_', ' ').title()
            fullname = fullname + ' ' + C.get('wording.%s' % engine)

        dboptions =  C.get('db.%s' % engine)
        if engine in ('mysqli', 'mariadb') and self.branch_compare(31):
            dboptions['charset'] = 'utf8mb4'

        logging.info('Creating database...')
        db = DB(engine, dboptions)
        if db.dbexists(dbname):
            if dropDb:
                db.dropdb(dbname)
                db.createdb(dbname)
            else:
                raise InstallException('Cannot install an instance on an existing database (%s)' % dbname)
        else:
            db.createdb(dbname)
        db.selectdb(dbname)

        logging.info('Installing %s...' % self.identifier)
        cli = 'admin/cli/install.php'
        params = (wwwroot, dataDir, engine, dbname, C.get('db.%s.user' % engine), C.get('db.%s.passwd' % engine), C.get('db.%s.host' % engine), C.get('db.%s.port' % engine), C.get('db.tablePrefix'), fullname, self.identifier, C.get('login'), C.get('passwd'))
        args = '--wwwroot="%s" --dataroot="%s" --dbtype="%s" --dbname="%s" --dbuser="%s" --dbpass="%s" --dbhost="%s" --dbport="%s" --prefix="%s" --fullname="%s" --shortname="%s" --adminuser="%s" --adminpass="%s" --allow-unstable --agree-license --non-interactive' % params
        result = self.cli(cli, args, stdout=None, stderr=None)
        if result[0] != 0:
            raise InstallException('Error while running the install, please manually fix the problem.\n- Command was: %s %s %s' % (C.get('php'), cli, args))

        configFile = os.path.join(self.path, 'config.php')
        os.chmod(configFile, 0666)
        try:
            if C.get('path') != '' and C.get('path') != None:
                self.addConfig('sessioncookiepath', '/%s/%s/' % (C.get('path'), self.identifier))
            else:
                self.addConfig('sessioncookiepath', '/%s/' % self.identifier)
        except Exception:
            logging.warning('Could not append $CFG->sessioncookiepath to config.php')

        # Add forced $CFG to the config.php if some are globally defined.
        forceCfg = C.get('forceCfg')
        if isinstance(forceCfg, dict):
            for cfgKey, cfgValue in forceCfg.iteritems():
                try:
                    logging.info('Setting up forced $CFG->%s to \'%s\' in config.php', cfgKey, cfgValue)
                    self.addConfig(cfgKey, cfgValue)
                except Exception:
                    logging.warning('Could not append $CFG->%s to config.php', cfgKey)

        self.reload()

    def isInstalled(self):
        """Returns whether this instance is installed or not"""
        # Reload the configuration if necessary.
        self._load()
        return self.installed == True

    @staticmethod
    def isInstance(path):
        """Check whether the path is a Moodle web directory"""
        version = os.path.join(path, 'version.php')
        try:
            f = open(version, 'r')
            lines = f.readlines()
            f.close()
        except:
            return False
        found = False
        for line in lines:
            if line.find('MOODLE VERSION INFORMATION') > -1:
                found = True
                break
        if not found:
            return False

        return True

    def isIntegration(self):
        """Returns whether an instance is an integration one or not"""
        r = C.get('upstreamRemote') or 'upstream'
        if not self.git().getRemote(r):
            r = 'origin'
        remote = self.git().getConfig('remote.%s.url' % r)
        if remote != None and remote.endswith('integration.git'):
            return True
        return False

    def isStable(self):
        """Assume an instance is stable if not integration"""
        return not self.isIntegration()

    def _load(self):
        """Loads the information"""
        if not self.isInstance(self.path):
            return False

        if self._loaded:
            return True

        # Extracts information from version.php
        self.version = {}
        version = os.path.join(self.path, 'version.php')
        if os.path.isfile(version):

            reVersion = re.compile(r'^\s*\$version\s*=\s*([0-9.]+)\s*;')
            reRelease = re.compile(r'^\s*\$release\s*=\s*(?P<brackets>[\'"])?(.+)(?P=brackets)\s*;')
            reMaturity = re.compile(r'^\s*\$maturity\s*=\s*([a-zA-Z0-9_]+)\s*;')
            reBranch = re.compile(r'^\s*\$branch\s*=\s*(?P<brackets>[\'"])?([0-9]+)(?P=brackets)\s*;')

            f = open(version, 'r')
            for line in f:
                if reVersion.search(line):
                    self.version['version'] = reVersion.search(line).group(1)
                elif reRelease.search(line):
                    self.version['release'] = reRelease.search(line).group(2)
                elif reMaturity.search(line):
                    self.version['maturity'] = reMaturity.search(line).group(1).replace('MATURITY_', '').lower()
                elif reBranch.search(line):
                    self.version['branch'] = reBranch.search(line).group(2)

            # Several checks about the branch
            try:
                # Has it been set?
                branch = self.version['branch']
            except:
                self.version['branch'] = self.version['release'].replace('.', '')[0:2]
                branch = self.version['branch']
            if int(branch) >= int(C.get('masterBranch')):
                self.version['branch'] = 'master'

            # Stable branch
            if self.version['branch'] == 'master':
                self.version['stablebranch'] = 'master'
            else:
                self.version['stablebranch'] = 'MOODLE_%s_STABLE' % self.version['branch']

            # Integration or stable?
            self.version['integration'] = self.isIntegration()

            f.close()
        else:
            # Should never happen
            raise Exception('This does not appear to be a Moodle instance')

        # Extracts parameters from config.php, does not handle params over multiple lines
        self.config = {}
        config = os.path.join(self.path, 'config.php')
        if os.path.isfile(config):
            self.installed = True
            prog = re.compile(r'^\s*\$CFG->([a-z_]+)\s*=\s*((?P<brackets>[\'"])?(.+)(?P=brackets)|([0-9.]+)|(true|false|null))\s*;$', re.I)
            try:
                f = open(config, 'r')
                for line in f:
                    match = prog.search(line)
                    if match == None:
                        continue

                    if match.group(5) != None:
                        # Number
                        value = float(match.group(5)) if '.' in str(match.group(5)) else int(match.group(5))
                    elif match.group(6) != None:
                        # Boolean or null
                        value = str(match.group(6)).lower()
                        if value == 'true':
                            value = True
                        elif value == 'false':
                            value = False
                        else:
                            value = None
                    else:
                        # Likely to be a string
                        value = match.group(4)

                    self.config[match.group(1)] = value

                f.close()

            except IOError:
                self.installed = False
                logging.error('Could not read config file')

        else:
            self.installed = False

        self._loaded = True
        return True

    def purge(self, manual=False):
        """Purge the cache of an instance"""
        if not self.isInstalled():
            raise Exception('Instance not installed, cannot purge.')
        elif self.branch_compare('22', '<'):
            raise Exception('Instance does not support cache purging.')

        try:
            dataroot = self.get('dataroot', False)
            if manual and dataroot != False:
                logging.debug('Removing directories [dataroot]/cache and [dataroot]/localcache')
                shutil.rmtree(os.path.join(dataroot, 'cache'), True)
                shutil.rmtree(os.path.join(dataroot, 'localcache'), True)

            self.cli('admin/cli/purge_caches.php', stderr=None, stdout=None)

        except Exception:
            raise Exception('Error while purging cache!')

    def pushPatch(self, branch=None):
        """Push a patch on the tracker, and remove the previous one"""

        if branch == None:
            branch = self.currentBranch()
            if branch == 'HEAD':
                raise Exception('Cannot create a patch from a detached branch')

        # Parsing the branch
        parsedbranch = parseBranch(branch)
        if not parsedbranch:
            raise Exception('Could not extract issue number from %s' % branch)
        issue = 'MDL-%s' % (parsedbranch['issue'])
        headcommit = self.headcommit(branch)

        # Creating a patch file.
        fileName = branch + '.mdk.patch'
        tmpPatchFile = os.path.join(gettempdir(), fileName)
        if self.git().createPatch('%s...%s' % (headcommit, branch), saveTo=tmpPatchFile):

            J = Jira()

            # Checking if file with same name exists.
            existingAttachmentId = None
            existingAttachments = J.getIssue(issue, fields='attachment')
            for existingAttachment in existingAttachments.get('fields', {}).get('attachment', {}):
                if existingAttachment.get('filename') == fileName:
                    # Found an existing attachment with the same name, we keep track of it.
                    existingAttachmentId = existingAttachment.get('id')
                    break

            # Pushing patch to the tracker.
            try:
                logging.info('Uploading %s to the tracker' % (fileName))
                J.upload(issue, tmpPatchFile)
            except JiraException:
                logging.error('Error while uploading the patch to the tracker')
                return False
            else:
                if existingAttachmentId != None:
                    # On success, deleting file that was there before.
                    try:
                        logging.info('Deleting older patch...')
                        J.deleteAttachment(existingAttachmentId)
                    except JiraException:
                        logging.info('Could not delete older attachment')

        else:
            logging.error('Could not create a patch file')
            return False

        return True

    def reload(self):
        """Sets the value to be reloaded"""
        self._loaded = False

    def removeConfig(self, name):
        """Remove a configuration setting from the config file."""
        configFile = os.path.join(self.path, 'config.php')
        if not os.path.isfile(configFile):
            return None

        try:
            f = open(configFile, 'r')
            lines = f.readlines()
            f.close()

            for line in lines:
                if re.search(r'\$CFG->%s\s*=.*;' % (name), line):
                    lines.remove(line)
                    break

            f = open(configFile, 'w')
            f.writelines(lines)
            f.close()
        except:
            raise Exception('Error while writing to config file')

        self.reload()

    def runScript(self, scriptname, arguments=None, **kwargs):
        """Runs a script on the instance"""
        return Scripts.run(scriptname, self.get('path'), arguments=arguments, cmdkwargs=kwargs)

    def update(self, remote=None):
        """Update the instance from the remote"""

        if remote == None:
            remote = C.get('upstreamRemote')

        # Fetch
        if not self.git().fetch(remote):
            raise Exception('Could not fetch remote %s' % remote)

        # Checkout stable
        self.checkout_stable(True)

        # Reset HARD
        upstream = '%s/%s' % (remote, self.get('stablebranch'))
        if not self.git().reset(to=upstream, hard=True):
            raise Exception('Error while executing git reset.')

        # Return to previous branch
        self.checkout_stable(False)

    def updateConfig(self, name, value):
        """Update a setting in the config file."""
        self.removeConfig(name)
        self.addConfig(name, value)

    def uninstall(self):
        """Uninstall the instance"""

        if not self.isInstalled():
            raise Exception('The instance is not installed')

        # Delete the content in moodledata
        dataroot = self.get('dataroot')
        if os.path.isdir(dataroot):
            logging.debug('Deleting dataroot content (%s)' % (dataroot))
            shutil.rmtree(dataroot)
            mkdir(dataroot, 0777)

        # Drop the database
        dbname = self.get('dbname')
        if self.dbo().dbexists(dbname):
            logging.debug('Droping database (%s)' % (dbname))
            self.dbo().dropdb(dbname)

        # Remove the config file
        configFile = os.path.join(self.get('path'), 'config.php')
        if os.path.isfile(configFile):
            logging.debug('Deleting config.php')
            os.remove(configFile)

    def updateTrackerGitInfo(self, branch=None, ref=None):
        """Updates the git info on the tracker issue"""

        if branch == None:
            branch = self.currentBranch()
            if branch == 'HEAD':
                raise Exception('Cannot update the tracker when on detached branch')

        # Parsing the branch
        parsedbranch = parseBranch(branch)
        if not parsedbranch:
            raise Exception('Could not extract issue number from %s' % branch)
        issue = 'MDL-%s' % (parsedbranch['issue'])
        version = parsedbranch['version']

        # Get the jira config
        repositoryurl = C.get('repositoryUrl')
        diffurltemplate = C.get('diffUrlTemplate')
        stablebranch = self.get('stablebranch')

        # Get the hash of the last upstream commit
        headcommit = None
        logging.info('Searching for the head commit...')
        if ref:
            try:
                headcommit = self.git().hashes(ref=ref, limit=1, format='%H')[0]
            except GitException:
                logging.warning('Could not resolve a head commit using the reference: %s' % (ref))
                headcommit = None

        # No reference was passed, or it was invalid.
        if not headcommit:
            headcommit = self.headcommit(branch)

        # Head commit not resolved
        if not headcommit:
            logging.error('Head commit not resolved, aborting update of tracker fields')
            return False

        headcommit = headcommit[:10]
        logging.debug('Head commit resolved to %s' % (headcommit))

        J = Jira()
        diffurl = diffurltemplate.replace('%branch%', branch).replace('%stablebranch%', stablebranch).replace('%headcommit%', headcommit)

        fieldrepositoryurl = C.get('tracker.fieldnames.repositoryurl')
        fieldbranch = C.get('tracker.fieldnames.%s.branch' % version)
        fielddiffurl = C.get('tracker.fieldnames.%s.diffurl' % version)

        if not fieldrepositoryurl or not fieldbranch or not fielddiffurl:
            logging.error('Cannot set tracker fields for this version (%s). The field names are not set in the config file.', version)
        else:
            logging.info('Setting tracker fields: \n  %s: %s \n  %s: %s \n  %s: %s' %
                (fieldrepositoryurl, repositoryurl, fieldbranch, branch, fielddiffurl, diffurl))
            J.setCustomFields(issue, {fieldrepositoryurl: repositoryurl, fieldbranch: branch, fielddiffurl: diffurl})

    def upgrade(self, nocheckout=False):
        """Calls the upgrade script"""
        if not self.isInstalled():
            raise Exception('Cannot upgrade an instance which is not installed.')
        elif not self.branch_compare(20):
            raise Exception('Upgrade command line tool not supported by this version.')
        elif os.path.isfile(os.path.join(self.get('path'), '.noupgrade')):
            raise UpgradeNotAllowed('Upgrade not allowed, found .noupgrade.')

        # Checkout stable
        if not nocheckout:
            self.checkout_stable(True)

        cli = '/admin/cli/upgrade.php'
        args = '--non-interactive --allow-unstable'
        result = self.cli(cli, args, stdout=None, stderr=None)
        if result[0] != 0:
            raise Exception('Error while running the upgrade.')

        # Return to previous branch
        if not nocheckout:
            self.checkout_stable(False)

    def uninstallPlugins(self, name):
        """Calls the CLI to uninstall a plugin"""

        if not self.branch_compare(37):
            raise Exception('Uninstalling plugins is only available from Moodle 3.7.')

        cli = '/admin/cli/uninstall_plugins.php'
        args = '--plugins=' + name + ' --run'

        result = self.cli(cli, args, stdout=subprocess.PIPE, stderr=None)
        try:
            result[1]
            resultstring = result[1].split('\n', 1)
            cannotuninstall = resultstring[0].rfind('Can not be uninstalled')
            if cannotuninstall != -1:
                raise Exception('The plugin could not be uninstalled')
        except IndexError as e:
            # We should always have some text returned and so should not end up here. And we
            # raise the exception as we're unsure whether the plugin was uninstalled properly
            # so it's better to halt the process.
            logging.error('The plugin uninstall cli code has changed and I need to be updated.')
            raise e



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import time
import logging
from distutils.errors import DistutilsFileError
from .. import backup
from ..command import Command
from ..exceptions import *


class BackupCommand(Command):

    _arguments = [
        (
            ['-i', '--info'],
            {
                'dest': 'info',
                'help': 'lists all the information about a backup',
                'metavar': 'backup'
            }
        ),
        (
            ['-l', '--list'],
            {
                'action': 'store_true',
                'dest': 'list',
                'help': 'list the backups'
            },
        ),
        (
            ['-r', '--restore'],
            {
                'dest': 'restore',
                'help': 'restore a backup',
                'metavar': 'backup'
            }
        ),
        (
            ['name'],
            {
                'default': None,
                'help': 'name of the instance',
                'nargs': '?'
            }
        )
    ]

    _description = 'Backup a Moodle instance'

    def run(self, args):
        name = args.name
        BackupManager = backup.BackupManager()

        # List the backups
        if args.list:
            backups = BackupManager.list()
            for key in sorted(backups.keys()):
                B = backups[key]
                backuptime = time.ctime(B.get('backup_time'))
                print '{0:<25}: {1:<30} {2}'.format(key, B.get('release'), backuptime)

        # Displays backup information
        elif args.info:
            name = args.info

            # Resolve the backup
            if not name or not BackupManager.exists(name):
                raise Exception('This is not a valid backup')

            # Restore process
            B = BackupManager.get(name)
            infos = B.infos
            print 'Displaying information about %s' % name
            for key in sorted(infos.keys()):
                print '{0:<20}: {1}'.format(key, infos[key])

        # Restore
        elif args.restore:
            name = args.restore

            # Resolve the backup
            if not name or not BackupManager.exists(name):
                raise Exception('This is not a valid backup')

            # Restore process
            B = BackupManager.get(name)

            try:
                M = B.restore()
            except BackupDirectoryExistsException:
                raise Exception('Cannot restore an instance on an existing directory. Please remove %s first.' % B.get('identifier') +
                    'Run: moodle remove %s' % B.get('identifier'))
            except BackupDBExistsException:
                raise Exception('The database %s already exists. Please remove it first.' % B.get('dbname') +
                    'This command could help: moodle remove %s' % B.get('identifier'))

            # Loads M object and display information
            logging.info('')
            logging.info('Restored instance information')
            logging.info('')
            infos = M.info()
            for key in sorted(infos.keys()):
                print '{0:<20}: {1}'.format(key, infos[key])
            logging.info('')

            logging.info('Done.')

        # Backup the instance
        else:
            M = self.Wp.resolve(name)
            if not M:
                raise Exception('This is not a Moodle instance')

            try:
                BackupManager.create(M)
            except BackupDBEngineNotSupported:
                raise Exception('Does not support backup for the DB engine %s yet, sorry!' % M.get('dbtype'))

            except DistutilsFileError:
                raise Exception('Error while copying files. Check the permissions on the data directory.' +
                    'Or run: sudo chmod -R 0777 %s' % M.get('dataroot'))

            logging.info('Done.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
from ..command import Command


class RemoveCommand(Command):

    _arguments = [
        (
            ['name'],
            {
                'help': 'name of the instance'
            }
        ),
        (
            ['-y'],
            {
                'action': 'store_true',
                'dest': 'do',
                'help': 'do not ask for confirmation'
            }
        ),
        (
            ['-f'],
            {
                'action': 'store_true',
                'dest': 'force',
                'help': 'force and do not ask for confirmation'
            }
        )
    ]
    _description = 'Completely remove an instance'

    def run(self, args):

        try:
            M = self.Wp.get(args.name)
        except:
            raise Exception('This is not a Moodle instance')

        if not args.do and not args.force:
            confirm = raw_input('Are you sure? (Y/n) ')
            if confirm != 'Y':
                logging.info('Aborting...')
                return

        logging.info('Removing %s...' % args.name)
        try:
            self.Wp.delete(args.name)
        except OSError:
            raise Exception('Error while deleting the instance.\n' +
                'This is probably a permission issue.\n' +
                'Run: sudo chmod -R 0777 %s' % self.Wp.getPath(args.name))

        logging.info('Instance removed')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
from ..command import Command


class RebaseCommand(Command):

    _description = 'Rebase branches'

    def __init__(self, *args, **kwargs):
        super(RebaseCommand, self).__init__(*args, **kwargs)
        self._arguments = [
            (
                ['-i', '--issues'],
                {
                    'help': 'issues to be rebased',
                    'metavar': 'issues',
                    'nargs': '+',
                    'required': True
                }
            ),
            (
                ['-s', '--suffix'],
                {
                    'help': 'the suffix of the branch of those issues',
                    'metavar': 'suffix'
                }
            ),
            (
                ['-v', '--versions'],
                {
                    'choices': [str(x) for x in range(13, int(self.C.get('masterBranch')))] + ['master'],
                    'help': 'versions to rebase the issues on. Ignored if names is set.',
                    'metavar': 'version',
                    'nargs': '+'
                }
            ),
            (
                ['-p', '--push'],
                {
                    'action': 'store_true',
                    'help': 'push the branch after successful rebase'
                }
            ),
            (
                ['-t', '--update-tracker'],
                {
                    'action': 'store_true',
                    'dest': 'updatetracker',
                    'help': 'to use with --push, also add the diff information to the tracker issue'
                }
            ),
            (
                ['-r', '--remote'],
                {
                    'default': self.C.get('myRemote'),
                    'help': 'the remote to push the branch to. Default is %s.' % self.C.get('myRemote'),
                    'metavar': 'remote'
                }
            ),
            (
                ['-f', '--force-push'],
                {
                    'action': 'store_true',
                    'dest': 'forcepush',
                    'help': 'Force the push'
                }
            ),
            (
                ['names'],
                {
                    'default': None,
                    'help': 'name of the instances to rebase',
                    'metavar': 'names',
                    'nargs': '*'
                }
            )
        ]

    def run(self, args):

        names = args.names
        issues = args.issues
        versions = args.versions

        # If we don't have a version, we need an instance
        if not names and not versions:
            raise Exception('This is not a Moodle instance')

        # We don't have any names but some versions are set
        if not names:
            names = []
            for v in versions:
                names.append(self.Wp.generateInstanceName(v))

        # Getting instances
        Mlist = self.Wp.resolveMultiple(names)

        # Updating cache remotes
        logging.info('Updating cached repositories')
        self.Wp.updateCachedClones(verbose=False)

        # Loops over instances to rebase
        for M in Mlist:
            logging.info('Working on %s' % (M.get('identifier')))
            M.git().fetch(self.C.get('upstreamRemote'))

            # Test if currently in a detached branch
            if M.git().currentBranch() == 'HEAD':
                result = M.git().checkout(M.get('stablebranch'))
                # If we can't checkout the stable branch, that is probably because we are in an unmerged situation
                if not result:
                    logging.warning('Error. The repository seem to be on a detached branch. Skipping.')
                    continue

            # Stash
            stash = M.git().stash(untracked=True)
            if stash[0] != 0:
                logging.error('Error while trying to stash your changes. Skipping %s.' % M.get('identifier'))
                logging.debug(stash[2])
                continue
            elif not stash[1].startswith('No local changes'):
                logging.info('Stashed your local changes')

            # Looping over each issue to rebase
            for issue in issues:
                branch = M.generateBranchName(issue, suffix=args.suffix)
                if not M.git().hasBranch(branch):
                    logging.warning('Could not find branch %s' % (branch))
                    continue

                # Rebase
                logging.info('> Rebasing %s...' % (branch))
                base = '%s/%s' % (self.C.get('upstreamRemote'), M.get('stablebranch'))
                result = M.git().rebase(branch=branch, base=base)
                if result[0] != 0:
                    logging.warning('Error while rebasing branch %s on top of %s' % (branch, base))
                    if result[0] == 1 and result[2].strip() == '':
                        logging.debug('There must be conflicts.')
                        logging.info('Aborting... Please rebase manually.')
                        M.git().rebase(abort=True)
                    else:
                        logging.debug(result[2])
                    continue

                # Pushing branch
                if args.push:
                    remote = args.remote
                    logging.info('Pushing %s to %s' % (branch, remote))
                    result = M.git().push(remote=remote, branch=branch, force=args.forcepush)
                    if result[0] != 0:
                        logging.warning('Error while pushing to remote')
                        logging.debug(result[2])
                        continue

                    # Update the tracker
                    if args.updatetracker:
                        M.updateTrackerGitInfo(branch=branch)

            # Stash pop
            if not stash[1].startswith('No local changes'):
                pop = M.git().stash(command='pop')
                if pop[0] != 0:
                    logging.error('An error occurred while unstashing your changes')
                    logging.debug(pop[2])
                else:
                    logging.info('Popped the stash')

            logging.info('')

        logging.info('Done.')
